/* ### Linear Feedback Shift Register (LFSR) ####
 */

#include "lfsr.h"
#include "uint48.h"

using namespace std;

LFSR::LFSR()
{
}

LFSR::~LFSR()
{
}

void LFSR::initialize(uint48 input)
{
	currentState = input;
}

/* The feedback function */
bool LFSR::L(uint48 input)
{
	bool result = input.getBit(0)^input.getBit(5)^input.getBit(9)^input.getBit(10)^input.getBit(12)^input.getBit(14)^
					input.getBit(15)^input.getBit(17)^input.getBit(19)^input.getBit(24)^input.getBit(25)^input.getBit(27)^
					input.getBit(29)^input.getBit(35)^input.getBit(39)^input.getBit(41)^input.getBit(42)^input.getBit(43);
	
	return result;
}

/* #### The filter function #### */
bool LFSR::f(uint48 input)
{
	bool fa1_input[4] = {input.getBit(9),input.getBit(11),input.getBit(13),input.getBit(15) };
	bool fa2_input[4] = {input.getBit(33),input.getBit(35),input.getBit(37),input.getBit(39) };
	bool fb1_input[4] = {input.getBit(17),input.getBit(19),input.getBit(21),input.getBit(23) };
	bool fb2_input[4] = {input.getBit(25),input.getBit(27),input.getBit(29),input.getBit(31) };
	bool fb3_input[4] = {input.getBit(41),input.getBit(43),input.getBit(45),input.getBit(47) };
	
	bool fc_input[5] = { fa(fa1_input),
					fb(fb1_input),
					fb(fb2_input),
					fa(fa2_input),
					fb(fb3_input) };
	bool result = fc(fc_input);
	return result;
}

/* Parts of the filter function */
bool LFSR::fa(bool input [4])
{
	bool result = ( (input[0] | input[1]) ^ (input[0] & input[3])) ^
					( input[2] & ((input[0] ^ input[1]) | input[3]) );
	return result;
}

bool LFSR::fb(bool input [4])
{
	bool result = ( (input[0] & input[1]) | input[2]) ^
					( (input[0]^input[1]) & (input[2] | input[3]) );
	return result;
}

bool LFSR::fc(bool input [5])
{
	bool result = (input[0] | ((input[1] | input[4]) & (input[3]^input[4]))) ^
					((input[0]^(input[1] & input[3])) & ((input[2]^input[3]) | (input[1] & input[4])));
	return result;
}
/* ############################## */


/* ###### Move the LFSR ###### */
/* For t > 64: Use only feedback function */
void LFSR::stepNext()
{
	bool newElement = L(currentState);
	currentState = currentState >> 1;
	currentState.setBit(newElement,47);
}
/* For t = [33..64]: Use the decrypted reader challenge and the feedback function as input */
void LFSR::stepNext(unsigned int uid, unsigned int challenge, int ticks)
{
	if(ticks < 32) {
		bool newElement = L(currentState) ^ (((uid^challenge) >> (31-ticks)) & 1);
		currentState = currentState >> 1;
		currentState.setBit(newElement,47);
	}
	else {
		stepNext();
	}
}
/* For t = [1..32]: Use Tag UID, Tag Challenge and feedback function (XORed) as input */
void LFSR::stepNext(unsigned int readerChallenge, int ticks)
{
	if(ticks < 64 && ticks >= 32) {
		bool newElement = L(currentState) ^ ((readerChallenge >> (63-ticks)) & 1);
		currentState = currentState >> 1;
		currentState.setBit(newElement,47);
	}
	else {
		stepNext();
	}
}
/* ############################## */


/* #### Rollback the LFSR #### */
/* For t > 64: Use only rollback function */
void LFSR::rollback(uint48 input)
{
	bool result = input.getBit(4)^input.getBit(8)^input.getBit(9)^input.getBit(11)^input.getBit(13)^input.getBit(14)^input.getBit(16)^input.getBit(18)^
					input.getBit(23)^input.getBit(24)^input.getBit(26)^input.getBit(28)^input.getBit(34)^input.getBit(38)^input.getBit(40)^input.getBit(41)^
					input.getBit(42)^input.getBit(47);
	currentState = (currentState << 1) | result;
}
/* For t = [33..64]: Use the decrypted reader challenge and the rollback function as input */
void LFSR::rollbackNRdecrypted(uint48 input, unsigned int readerChallenge, int ticks)
{
	rollback(input);
	if(ticks <= 64 && ticks > 32) {
		currentState = currentState^((readerChallenge >> (64-ticks)) & 1);
	}
}
/* For t = [33..64]: Use the encrypted reader challenge and the rollback function as input */
void LFSR::rollbackNRencrypted(uint48 input, unsigned int readerChallenge, int ticks)
{
	rollback(input);
	if(ticks <= 64 && ticks > 32) {
		bool fBit = f(input << 1);
		currentState = currentState^((readerChallenge >> (64-ticks)) & 1)^fBit;
	}
}
/* For t = [1..32]: Use Tag UID, Tag Challenge and rollback function (XORed) as input */
void LFSR::rollbackUID(uint48 input, unsigned int uid, unsigned int challenge, int ticks)
{
	rollback(input);
	if(ticks <= 32 && ticks > 0) {
		currentState = currentState^( ((uid^challenge) >> (32-ticks)) & 1 );
	}
}
/* ############################## */



bool LFSR::getBit(unsigned int index)
{
	return currentState.getBit(index);
}

uint48 LFSR::getState()
{
	return currentState;
}

bool LFSR::getKeystream()
{
	bool result = f(currentState);
	return result;
}
