/* ##### Simulation of a Mifare Classic proximity card ##### */


#include "mifaretag.h"
#include "lfsr.h"
#include "randomgenerator.h"
#include "tagsector.h"
#include "uint48.h"
#include "helpfunctions.h"
#include <iostream>
#include <iomanip>
#include <time.h>
#include <math.h>
#include <fstream>

using namespace std;

// Switch on/off debug information
bool showLFSRstate = false;

/* ##### Constructors and destructors ###### */
MifareTag::MifareTag()
{
	MifareTag(1);
}

MifareTag::MifareTag(unsigned int _sectorCount)
{
	mLFSR = new LFSR();
	mRandomGenerator = new RandomGenerator();
	sectorCount = _sectorCount;
	sectors = new TagSector[sectorCount];
	ticks = 0;
	uid = 0xc108416a;
	challenge = 0;
	readerChallenge = 0;
}

MifareTag::~MifareTag()
{
	delete mLFSR;
	delete mRandomGenerator;
}
/* ########################### */



/* ##### Getters and setters ###### */
uint48 MifareTag::getLFSRState()
{
	return mLFSR->getState();
}

unsigned int MifareTag::getRandomGeneratorLFSRState()
{
	return mRandomGenerator->getLFSRState();
}

unsigned int MifareTag::getRandomGeneratorNonceState()
{
	return mRandomGenerator->getNonceState();
}

unsigned int MifareTag::getUID()
{
	return uid;
}

bool MifareTag::getKeystream()
{
	return mLFSR->getKeystream();
}

void MifareTag::resetTag()
{
	ticks = 0;
}

unsigned int MifareTag::getTicks()
{
	return ticks;
}

unsigned int MifareTag::getChallenge(unsigned int sectorIndex)
{

	if(sectorIndex >= sectorCount) {
		return 0x05;
	}
	uint48 LFSRstate;
	unsigned int sucState;
	
	// Initialize LFSR
	uint48 sectorKey = sectors[sectorIndex].getKey();
	mLFSR->initialize(sectorKey);
	// Get challenge from RandomGenereator
	sucState = mRandomGenerator->getNonceState();
	sucState = mRandomGenerator->calculateSuc(sucState);
	challenge = sucState;
	
	// Advance time:
	for(int t=1;t<=32;t++) {
		if(showLFSRstate) {
			LFSRstate = mLFSR->getState();
			cout << ticks << ":\tLFSRState: ";
			cout << std::hex << LFSRstate.getHighInt() << LFSRstate.getLowInt() << endl;
		}
		tick();
	}
	
	return challenge;
}
/* ########################### */

/* Advance time */
void MifareTag::tick()
{
	if(ticks < 0)
	{
		ticks = 0;
	}
	if(ticks < 64)
	{
		if(ticks < 32) {
			mLFSR->stepNext(uid, challenge, ticks);
		}
		else {
			mLFSR->stepNext(readerChallenge, ticks); // Decrypted readerChallenge!
		}
	}
	else {
		mLFSR->stepNext();
	}
	mRandomGenerator->stepNext();
	ticks++;
}

/* Answer the reader challenge and reader answer
 * The actual answer, if all informations are valid, is not implemented
 * If the parity bits are right the encoded error message is returned
 */
unsigned int MifareTag::answerChallenge(unsigned int _readerChallenge, unsigned int _readerAnswer, unsigned int parityBits)
{
	unsigned int decodedReaderChallenge = _readerChallenge;
	unsigned int decodedReaderAnswer = _readerAnswer;
	
	unsigned int parityCounter = 0;
	unsigned int bytesProcessed = 0;
	bool parityWrong = false;
	bool parityExpected[8];
	uint48 LFSRstate;
	uint48 maskedLFSRstate;
	uint48 mask; mask.setHighInt(0xaaaa); mask.setLowInt(0xaaaaaaaa);

	// Process readerChallenge:
	for(int t=32;t<=64;t++) {
		bool b_t = mLFSR->getKeystream();
		if(t < 64) {
			decodedReaderChallenge ^= (b_t << (63-t));
			readerChallenge = decodedReaderChallenge;
		}
		
		if(t != 32 && t % 8 == 0) {
			bytesProcessed++;
			int shift = 32 - 8*(bytesProcessed);
			unsigned char byte = (decodedReaderChallenge >> shift) & 0xff;
			if(showLFSRstate) cout << std::hex << (unsigned int)byte << " => ";
			parityExpected[parityCounter] = calculateParityBit(byte)^b_t;
			if(showLFSRstate) cout << parityExpected << endl;
			parityCounter++;
			bool parityReceived = (parityBits >> (8-(parityCounter))) & 0x1;
			if(parityExpected[parityCounter-1] != (parityReceived)) {
				parityWrong = true;
			}
		}
		if(t < 64) {
			if(showLFSRstate) {
				LFSRstate = mLFSR->getState();
				maskedLFSRstate = LFSRstate & mask;
				cout << ticks << ":\tLFSRState: ";
				cout << std::hex << LFSRstate.getHighInt() << LFSRstate.getLowInt() << " - " <<  maskedLFSRstate.getHighInt() << maskedLFSRstate.getLowInt() << " ::: ";
				cout << _readerChallenge << " ::: " << decodedReaderChallenge << " - " << b_t << endl;
			}
			tick();
		}
	}
	// Process readerAnswer:
	bytesProcessed = 0;
	for(int t=64;t<=96;t++) {
		bool b_t = mLFSR->getKeystream();
		if(t < 96) {
			decodedReaderAnswer ^= (b_t << (95-t));
			readerAnswer = decodedReaderAnswer;
		}
		
		if(t != 64 && t % 8 == 0) {
			bytesProcessed++;
			int shift = 32 - 8*(bytesProcessed);
			unsigned char byte = (decodedReaderAnswer >> shift) & 0xff;
			if(showLFSRstate) cout << std::hex << (unsigned int)byte << " - " << b_t << " => ";
			parityExpected[parityCounter] = calculateParityBit(byte)^b_t;
			if(showLFSRstate) cout << parityExpected[parityCounter] << endl;
			parityCounter++;
			bool parityReceived = (parityBits >> (8-(parityCounter))) & 0x1;
			if(parityExpected[parityCounter-1] != (parityReceived)) {
				parityWrong = true;
			}
		}
		if(t < 96) {
			if(showLFSRstate) {
				LFSRstate = mLFSR->getState();
				maskedLFSRstate = LFSRstate & mask;
				cout << ticks << ":\tLFSRState: ";
				cout << std::hex << LFSRstate.getHighInt() << LFSRstate.getLowInt() << " - " <<  maskedLFSRstate.getHighInt() << maskedLFSRstate.getLowInt() << endl;
			}
			tick();
		}
	}
	
	if(parityWrong) {
		if(showLFSRstate) {
			cout << " - Expected: ";
			unsigned int exP = 0;
			for(int i=0;i<8;i++) {
				exP |= parityExpected[i] << (7-i);
			}
			cout << std::hex << exP << endl;
		}
		return 0xffffabcd;
	}
	else
	{
		// TODO: Check readerAnswer
		
		
		unsigned int encodedError = 0x5;
		for(int t=96;t<=99;t++) {
			bool b_t = mLFSR->getKeystream();
			encodedError ^= (b_t << (99-t));
			if(showLFSRstate) {
				LFSRstate = mLFSR->getState();
				cout << ticks << ":\tLFSRState: ";
				cout << std::hex << LFSRstate.getHighInt() << LFSRstate.getLowInt() << endl;
			}
			tick();
		}
		return encodedError;
	}
}

/* Rollback LFSR state to time = 0, i.e. return the sector key */
uint48 MifareTag::recoverKeyFromState(uint48 LFSRstate)
{
	cout << endl << "Starting rollback..." << endl;
	while(true)
	{
		LFSRstate = mLFSR->getState();
		if(ticks % 65535 == 0 || ticks <= 64) {
			cout << ticks << ":\tLFSRState: ";
			cout << std::hex << LFSRstate.getHighInt() << LFSRstate.getLowInt() << endl;
		}
		if(ticks > 64) {
			mLFSR->rollback(LFSRstate);
		}
		else {
			if(ticks >= 32) {
				mLFSR->rollbackNRdecrypted(LFSRstate,readerChallenge,ticks);
			}
			else {
				mLFSR->rollbackUID(LFSRstate,uid,challenge,ticks);
			}
		}
		if(ticks == 0) {
			break;
		}
		ticks--;
	}
	return LFSRstate;
}
